package offerCC

import (
	"github.com/pkg/errors"
	"gitlab.smartblocklab.com/smartcontracts/common/chaincode"
	"gitlab.smartblocklab.com/smartcontracts/common/remote/remote_account"
	"gitlab.smartblocklab.com/smartcontracts/common/invoke"
	"encoding/json"
	"reflect"
	"gitlab.smartblocklab.com/smartcontracts/common/model/accountCC"
	"gitlab.smartblocklab.com/smartcontracts/common/nums"
	"gitlab.smartblocklab.com/smartcontracts/common/model/transactionCC"
)

// OfferRecord keeps information relevant for user offering request
type OfferRecord struct {
	// same as TransactionRecord
	Key             string                        `json:"key"`
	OperationID     string                        `json:"operation_id,omitempty"`
	SessionID       string                        `json:"session_id,omitempty"`
	UserAccountInfo *accountCC.UserAccountInfo    `json:"user_account_info,omitempty"`
	Amount          nums.PlainCurrencyNumber      `json:"amount,omitempty"`
	HashCode        string                        `json:"hash_code,omitempty"`
	DependencyKey   string                        `json:"dependency_key,omitempty"` // transaction dependency in a form of "txnType.currencyType.accountId"
	Timestamp       transactionCC.Timestamp       `json:"timestamp"`
	InOutType       transactionCC.InputOutputType `json:"input_output_type"`
	// specific offer record
	ReplayedKey       string            `json:"replayed_key"`
	OfferRecordStatus OfferRecordStatus `json:"offer_record_status,omitempty"`
}

func (this OfferRecord) GetObjectType() string {
	return "OfferRecord";
}

const ExternalTransactionID = "external_transaction_id"

func (this OfferRecord) CheckBeforeAdd(manager chaincode.IManager) error {
	if len(this.ReplayedKey) != 0 {
		bytes, err := manager.GetState(this.ReplayedKey)
		if err != nil {
			return err
		}

		_, err = Unmarshal(bytes)
		if err != nil {
			return err
		}
	}

	// check out user account status
	accountRemoteManager := remote_account.GetAccountRemoteManager(manager, invoke.AccountChannel, invoke.AccountCC)

	account, err := accountRemoteManager.VerifyUserAccount(this.UserAccountInfo);
	if err != nil || account != nil {
		return err
	}

	return nil
}

func (this OfferRecord) CheckBeforeUpdate(updatedObject chaincode.ValidObject, manager chaincode.IManager) error {
	return errors.New("METHOD_DOES_NOT_IMPLEMENT")
}

func (this OfferRecord) GetKey(manager chaincode.IManager) []string {
	if len(this.SessionID) != 0 {
		return []string{this.SessionID, this.UserAccountInfo.UserID, this.UserAccountInfo.WalletID, this.UserAccountInfo.AccountID, this.Amount.CurrencyType, this.InOutType.String(), this.OperationID, this.OfferRecordStatus.String(), manager.GetTxID()}
	}
	return []string{this.UserAccountInfo.UserID, this.UserAccountInfo.WalletID, this.UserAccountInfo.AccountID, this.Amount.CurrencyType, this.InOutType.String(), this.OperationID, this.OfferRecordStatus.String(), manager.GetTxID()}
}

func (this OfferRecord) New() chaincode.ValidObject {
	return &OfferRecord{}
}

func (this OfferRecord) Validate() error {
	if len(this.OperationID) == 0 {
		return errors.New("OperationID is empty")
	}

	if len(this.OfferRecordStatus) == 0 {
		return errors.New("RequestStatus is empty")
	}

	if (len(this.ReplayedKey) == 0 && this.OfferRecordStatus == OfferRecordStatusReplay) ||
		(len(this.ReplayedKey) != 0 && this.OfferRecordStatus != OfferRecordStatusReplay) {
		return errors.Errorf("Error. Only OfferRecord with ReplayedKey need to have OfferRecordStatus OfferRecordStatus_REPLAY")
	}

	if this.UserAccountInfo == nil {
		return errors.New("UserAccountInfo is empty")
	}

	err := this.UserAccountInfo.Validate()
	if err != nil {
		return errors.Errorf("UserAccountInfo.%s", err.Error())
	}

	err = this.Amount.Validate()
	if err != nil {
		return err
	}

	isAmountEmpty, err := this.Amount.IsAmountEmpty()
	if err != nil {
		return err
	}
	if isAmountEmpty {
		return errors.New("Amount in OfferRecord can't be 0")
	}

	return nil
}

func Unmarshal(data []byte) (*OfferRecord, error) {
	if data == nil {
		return nil, chaincode.GlobalLogger.Error("can't Unmarshal nil data")
	}

	var obj OfferRecord
	err := json.Unmarshal(data, &obj)
	return &obj, err;
}

func Marshal(data *OfferRecord) ([]byte, error) {
	return json.Marshal(data);
}

func Cast(validObject chaincode.ValidObject) (*OfferRecord, error) {
	if validObject == nil {
		return nil, nil
	}

	if reflect.ValueOf(validObject).Kind() != reflect.Ptr {
		return nil, errors.New("check valid object is not reference")
	}

	casted, ok := validObject.(*OfferRecord)
	if !ok {
		return nil, errors.New("Problem to get object")
	}

	return casted, nil;
}
