package transactionCC

import (
	"encoding/json"
	"errors"
	"fmt"
	"gitlab.smartblocklab.com/smartcontracts/common/chaincode"
	"gitlab.smartblocklab.com/smartcontracts/common/invoke"
	"gitlab.smartblocklab.com/smartcontracts/common/model/accountCC"
	"gitlab.smartblocklab.com/smartcontracts/common/model/preferenceCC"
	"gitlab.smartblocklab.com/smartcontracts/common/model/requests"
	"gitlab.smartblocklab.com/smartcontracts/common/model/systempreferenceCC"
	"gitlab.smartblocklab.com/smartcontracts/common/model/userCC"
	"gitlab.smartblocklab.com/smartcontracts/common/nums"
	"gitlab.smartblocklab.com/smartcontracts/common/remote/remote_account"
	"gitlab.smartblocklab.com/smartcontracts/common/remote/remote_systempreference"
	"gitlab.smartblocklab.com/smartcontracts/common/remote/remote_user"
	"reflect"
	"strconv"
)

// TransactionRecord transaction struct which denotes transaction
// fields which are necessary for processing of
// transactions and will be used to present transaction in the ledger
type TransactionRecord struct {
	Key                        string                     `json:"key"`
	OperationID                string                     `json:"operation_id,omitempty"`
	ProcessType                ProcessType                `json:"process_type,omitempty"`
	OperationType              OperationType              `json:"operation_type,omitempty"`
	UserAccountInfo            *accountCC.UserAccountInfo `json:"user_account_info,omitempty"`
	DestinationUserAccountInfo *accountCC.UserAccountInfo `json:"destination_user_account_info,omitempty"`
	Amount                     nums.PlainCurrencyNumber   `json:"amount,omitempty"`
	HashCode                   string                     `json:"hash_code,omitempty"`
	DependencyKey              string                     `json:"dependency_key,omitempty"` // transaction dependency in a form of "txnType.currencyType.accountId"
	Timestamp                  Timestamp                  `json:"timestamp"`
	InOutType                  InputOutputType            `json:"input_output_type"`
	FromUserAccountInfo        *accountCC.UserAccountInfo `json:"from_user_account_info,omitempty"` // Информационное поле на него нельзя завязывать никакой бизнеслогики
	ToUserAccountInfo          *accountCC.UserAccountInfo `json:"to_user_account_info,omitempty"` // Информационное поле на него нельзя завязывать никакой бизнеслогики
}

func (this TransactionRecord) GetObjectType() string {
	return "TransactionRecord";
}

func (this TransactionRecord) CheckBeforeAdd(manager chaincode.IManager) error {

	// проверка что пользователь не заблокирован
	userRemoteManager := remote_user.GetUserRemoteManager(manager, invoke.UserChannel, invoke.UserCC)

	user, err := userRemoteManager.GetUser(&requests.Id{this.UserAccountInfo.UserID})
	if err != nil {
		return err
	}

	if user.UserStatus != userCC.UserStatus_Activated {
		return chaincode.GlobalLogger.Errorf("user %s is not ACTIVATED", user.UserID)
	}

	// проверка что в канале транзакций только записи с внутренними счетами и кошельками в статусе ACTIVE
	accountRemoteManager := remote_account.GetAccountRemoteManager(manager, invoke.AccountChannel, invoke.AccountCC)

	walletKeysWithFilter := &requests.KeysWithFilter{}
	walletKeysWithFilter.Keys = []string{this.UserAccountInfo.UserID, this.UserAccountInfo.WalletID}
	wallet, err := accountRemoteManager.GetAccount(walletKeysWithFilter)
	if err != nil {
		return err
	}

	if wallet.Status != accountCC.AccountStatus_ACTIVE {
		return chaincode.GlobalLogger.Errorf("wallet %s is not ACTIVE", wallet.WalletID)
	}

	accountKeysWithFilter := &requests.KeysWithFilter{}
	accountKeysWithFilter.Keys = []string{this.UserAccountInfo.UserID, this.UserAccountInfo.WalletID, this.UserAccountInfo.AccountID}
	account, err := accountRemoteManager.GetAccount(accountKeysWithFilter)
	if err != nil {
		return err
	}

	if account.Status != accountCC.AccountStatus_ACTIVE {
		return chaincode.GlobalLogger.Errorf("account %s is not ACTIVE", account.WalletID)
	}

	if account.CurrencyType != accountCC.INTERNAL_CURRENCY_TYPE {
		return chaincode.GlobalLogger.Errorf("account %s is not internal (ANY)", account.WalletID)
	}

	// TODO добавить проверку в регуляторном канале передав this
	// проверка статических переменных
	// общие данные в общем канале
	// время добавления
	// кто добавления
	// время обновления
	// кто обновил

	// проверка типа процесса и доступных криптовалют для этого процесса
	processTypeCurrencyRemoteManager := remote_systempreference.GetSystempreferenceRemoteManager(manager, invoke.ProcesstypecurrencyChannel, invoke.ProcesstypecurrencyCC)

	// TODO
	// userId Organization - user
	// ProcessType
	// CurrencyType
	// fields.put(CurrencyConstraintFields.IS_ENABLED, String.valueOf(isEnabled));
	// fields.put(CurrencyConstraintFields.MAX_VALUE, currencyConstraint.getMaxValue().toPlainString());
	// fields.put(CurrencyConstraintFields.MIN_VALUE, currencyConstraint.getMinValue().toPlainString());
	// fields.put(CurrencyConstraintFields.SCALE, String.valueOf(currencyConstraint.getScale()));
	processTypeKeyWithFilter := &requests.KeysWithFilter{}
	processTypeKeyWithFilter.Keys = []string{user.PartnerCode, this.ProcessType.String(), this.Amount.CurrencyType}
	processTypePreferenceArray, err := processTypeCurrencyRemoteManager.GetPreferenceArrayByKeysWithFilter(processTypeKeyWithFilter)
	if err != nil {
		return err
	}
	if processTypePreferenceArray == nil || len(processTypePreferenceArray.Preference) == 0 {
		return chaincode.GlobalLogger.Errorf("Process type %s not found", this.ProcessType.String())
	}

	if len(processTypePreferenceArray.Preference) != 1 {
		return chaincode.GlobalLogger.Errorf("Process type found %s expected %s", len(processTypePreferenceArray.Preference), 1)
	}

	processTypePreference := processTypePreferenceArray.Preference[0]

	chaincode.GlobalLogger.Infof("processTypePreference %s", processTypePreference)

	err = this.CheckAsset(&processTypePreference)
	if err != nil {
		return err
	}

	isAmountEmpty, err := this.Amount.IsAmountEmpty()
	if err != nil {
		return err
	}
	if isAmountEmpty {
		return chaincode.GlobalLogger.Error("amount can't be 0")
	}

	// проверка типа криптовалюты
	systemPreferenceRemoteManager := remote_systempreference.GetSystempreferenceRemoteManager(manager, invoke.CurrencyChannel, invoke.CurrencyCC)
	cryptocurrencyConstraintPreference, err := systemPreferenceRemoteManager.GetPreference(&requests.Id{Id: this.Amount.CurrencyType})
	if err != nil {
		return err
	}
	if cryptocurrencyConstraintPreference == nil {
		return chaincode.GlobalLogger.Errorf("Constraints for CurrencyType %s not found", this.Amount.CurrencyType)
	}

	err = this.CheckAsset(cryptocurrencyConstraintPreference)
	if err != nil {
		return err
	}

	return nil
}

func (this TransactionRecord) CheckAsset(constraintPreference *preferenceCC.Preference) error {
	if constraintPreference.Fields[systempreferenceCC.IS_ENABLED] != "true" {
		return chaincode.GlobalLogger.Errorf("Currency %s is not enabled", constraintPreference.Fields[systempreferenceCC.CURRENCY_TYPE])
	}

	scaleUint64, err := strconv.ParseUint(constraintPreference.Fields[systempreferenceCC.SCALE], 10, 32)
	if err != nil {
		return chaincode.GlobalLogger.Error(err.Error())
	}
	scale := uint(scaleUint64)

	newAsset, err := nums.NewAsset(constraintPreference.Fields[systempreferenceCC.CURRENCY_TYPE], scale, constraintPreference.Fields[systempreferenceCC.MIN_VALUE], constraintPreference.Fields[systempreferenceCC.MAX_VALUE])
	number, err := this.Amount.Number()
	if err != nil {
		return chaincode.GlobalLogger.Error(err.Error())
	}

	number = number.UpdateAsset(newAsset)
	_, err = number.Validate()
	if err != nil {
		return chaincode.GlobalLogger.Error(err.Error())
	}

	err = this.Amount.ValidateAmount()
	if err != nil {
		return chaincode.GlobalLogger.Error(err.Error())
	}
	return nil
}

func (this TransactionRecord) CheckBeforeUpdate(updatedObject chaincode.ValidObject, manager chaincode.IManager) error {
	return errors.New("METHOD_DOES_NOT_IMPLEMENT")
}

func (this TransactionRecord) GetKey(manager chaincode.IManager) []string {
	return []string{this.UserAccountInfo.UserID, this.UserAccountInfo.WalletID, this.UserAccountInfo.AccountID, this.Amount.CurrencyType, this.InOutType.String(), this.OperationID, this.OperationType.String(), manager.GetTxID()}
}

func (this TransactionRecord) New() chaincode.ValidObject {
	return &TransactionRecord{}
}

func (this TransactionRecord) Validate() error {

	if len(this.DependencyKey) == 0 {
		return errors.New("DependencyKey is empty")
	}

	if this.UserAccountInfo == nil {
		return errors.New("UserAccountInfo is empty")
	}

	if len(this.UserAccountInfo.UserID) == 0 {
		return errors.New("UserID is empty")
	}

	if len(this.Amount.CurrencyType) == 0 {
		return errors.New("CurrencyType is empty")
	}

	if len(this.OperationID) == 0 {
		return errors.New("OperationID is empty")
	}

	if len(this.InOutType) == 0 {
		return errors.New("InOutType is empty")
	}

	if this.UserAccountInfo == nil {
		errMsg := fmt.Sprintf("UserAccountInfo can't be nil")
		return errors.New(errMsg)
	}

	err := this.UserAccountInfo.Validate()
	if err != nil {
		return err
	}

	// для входяших транзакций к пользователю на блокировку нужно делать проверку для поля DestinationUserAccountInfo
	// для остальных операций этого поля НЕТ
	if this.OperationType == LockAmount && this.InOutType == INPUT {

		if this.DestinationUserAccountInfo == nil {
			return errors.New("DestinationUserAccountInfo can't be nil")
		}

		err = this.DestinationUserAccountInfo.Validate()
		if err != nil {
			return errors.New("DestinationUserAccountInfo. " + err.Error())
		}
	}

	err = this.Amount.Validate()
	if err != nil {
		return err
	}

	return nil
}

func Unmarshal(data []byte) (*TransactionRecord, error) {
	if data == nil {
		return nil, chaincode.GlobalLogger.Error("can't Unmarshal nil data")
	}

	var obj TransactionRecord
	err := json.Unmarshal(data, &obj)
	return &obj, err;
}

func Marshal(data *TransactionRecord) ([]byte, error) {
	return json.Marshal(data);
}

func Cast(validObject chaincode.ValidObject) (*TransactionRecord, error) {
	if validObject == nil {
		return nil, nil
	}

	if reflect.ValueOf(validObject).Kind() != reflect.Ptr {
		return nil, errors.New("check valid object is not reference")
	}

	casted, ok := validObject.(*TransactionRecord)
	if !ok {
		return nil, errors.New("Problem to get object")
	}

	return casted, nil;
}
