package transactionCC

import (
	"github.com/pkg/errors"
	//"encoding/json"
	"gitlab.smartblocklab.com/smartcontracts/common/nums"

	"encoding/json"
	"gitlab.smartblocklab.com/smartcontracts/common/chaincode"
	"gitlab.smartblocklab.com/smartcontracts/common/logging"
	"gitlab.smartblocklab.com/smartcontracts/common/model/accountCC"
)

type (
	// txnProcessCommissionType
	TxnProcessCommissionType string
)

// TxnFundsLockInput encapsulates input parameters
// for lock funds transaction type
type OutTxnFundsLockInput struct {
	OperationID string                     `json:"operation_id,omitempty"`
	OutUser     *accountCC.UserAccountInfo `json:"out_user,omitempty"`
	ProcessType ProcessType                `json:"process_type,omitempty"`
	Amount      nums.PlainCurrencyNumber   `json:"amount,omitempty"`
}

type Transfer struct {
	OperationID     string                    `json:"operation_id,omitempty"`
	UserAccountInfo accountCC.UserAccountInfo `json:"user_account_info,omitempty"`
}

func Unmarshal_Transfer(data []byte) (*Transfer, error) {
	if data == nil {
		return nil, logging.Get().Error("can't Unmarshal nil data")
	}

	var obj Transfer
	err := json.Unmarshal(data, &obj)
	return &obj, err
}

func (this Transfer) Marshal() ([]byte, error) {
	return json.Marshal(this)
}

type LockFundsOnAccountArg struct {
	OperationID string `json:"operation_id,omitempty"`
}

// TxnFundsLockInput encapsulates input parameters
// for lock funds transaction type
type TxnFundsLockInput struct {
	OperationID string                     `json:"operation_id,omitempty"`
	OutUser     *accountCC.UserAccountInfo `json:"out_user,omitempty"`
	InUser      *accountCC.UserAccountInfo `json:"in_user,omitempty"`
	ProcessType ProcessType                `json:"process_type,omitempty"`
	Amount      nums.PlainCurrencyNumber   `json:"amount,omitempty"`
}

//FundsTransferInput input for funds transfer transaction
type FundsTransferInput struct {
	Txns []TxnFundsLockInput `json:"transactions"`
}

type TransferFundsResponseList struct {
	ResponseList []TransferFundsResponse `json:"transfer_funds_response_list"`
}

type TransferFundsResponse struct {
	TransactionRecordList []*TransactionRecord `json:"transaction_record_list,omitempty"`
	ErrorMsg              string               `json:"error_msg,omitempty"`
	Status                ResultStatus         `json:"status,omitempty"`
}

// Save persist transaction record into ledger state
// TransactionList captures all user's transactions into
// struct to be able to encapsulate query result and return
// back to the invoker
type TransactionList struct {
	Transactions []*TransactionRecord `json:"transactions"`
}

type UnlockTransactionInfoList struct {
	LockTransactionRecordInfoList []LockTransactionRecordInfo
}

// проверка что все транзакции из списка имеют один ProcessType и operationID
// при условии что для типа процесса "Транзакция комиссия с получателя" комиссия не указана
func (this *UnlockTransactionInfoList) Validate() error {
	for _, lockTransactionRecordInfo := range this.LockTransactionRecordInfoList {
		err := lockTransactionRecordInfo.ValidateOperationID()
		if err != nil {
			return err
		}
		err = lockTransactionRecordInfo.ValidateProcessType()
		if err != nil {
			return err
		}
	}

	return nil
}

type LockTransactionRecordInfo struct {
	InLockAmountTransactionRecord      *TransactionRecord
	OutLockAmountTransactionRecord     *TransactionRecord
	InLockCommissionTransactionRecord  *TransactionRecord
	OutLockCommissionTransactionRecord *TransactionRecord
}

// проверка что OperationID совпадает у всех транзакций
// поля InLockCommissionTransactionRecord и OutLockCommissionTransactionRecord могут быть nil для процесса TransferReceiverCommission_ProcessType
func (this LockTransactionRecordInfo) ValidateOperationID() error {
	if this.InLockAmountTransactionRecord == nil || len(this.InLockAmountTransactionRecord.OperationID) == 0 {
		return errors.Errorf("InLockAmountTransactionRecord need to be not nil and have field OperationID")
	}
	if this.OutLockAmountTransactionRecord == nil || len(this.OutLockAmountTransactionRecord.OperationID) == 0 {
		return errors.Errorf("InLockAmountTransactionRecord need to be not nil and have field OperationID")
	}

	if this.InLockAmountTransactionRecord.OperationID != this.OutLockAmountTransactionRecord.OperationID {
		return errors.Errorf("Operation id not equals")
	}

	if this.InLockCommissionTransactionRecord != nil &&
		this.InLockAmountTransactionRecord.OperationID != this.InLockCommissionTransactionRecord.OperationID {
		return errors.Errorf("Operation id not equals")
	}

	if this.OutLockCommissionTransactionRecord != nil &&
		this.InLockAmountTransactionRecord.OperationID != this.OutLockCommissionTransactionRecord.OperationID {
		return errors.Errorf("Operation id not equals")
	}

	return nil
}

// проверка что OperationID совпадает
// поля InLockCommissionTransactionRecord и OutLockCommissionTransactionRecord могут быть nil для процесса TransferReceiverCommission_ProcessType
func (this LockTransactionRecordInfo) ValidateProcessType() error {
	if this.InLockAmountTransactionRecord != nil && this.OutLockAmountTransactionRecord != nil &&
		this.InLockAmountTransactionRecord.ProcessType != this.OutLockAmountTransactionRecord.ProcessType {
		return errors.Errorf("ProcessType not equals InLockAmountTransactionRecord %s and OutLockAmountTransactionRecord %s",
			this.InLockAmountTransactionRecord.ProcessType,
			this.OutLockAmountTransactionRecord.ProcessType)
	}

	if this.InLockCommissionTransactionRecord != nil && this.OutLockCommissionTransactionRecord != nil &&
		this.InLockCommissionTransactionRecord.ProcessType != this.OutLockCommissionTransactionRecord.ProcessType {
		return errors.Errorf("ProcessType not equals InLockCommissionTransactionRecord %s and OutLockCommissionTransactionRecord %s",
			this.InLockCommissionTransactionRecord.ProcessType,
			this.OutLockCommissionTransactionRecord.ProcessType)
	}

	if this.InLockCommissionTransactionRecord != nil && this.InLockAmountTransactionRecord != nil &&
		this.InLockCommissionTransactionRecord.ProcessType != this.InLockAmountTransactionRecord.ProcessType {
		return errors.Errorf("ProcessType not equals InLockCommissionTransactionRecord %s and InLockAmountTransactionRecord %s",
			this.InLockCommissionTransactionRecord.ProcessType,
			this.InLockAmountTransactionRecord.ProcessType)
	}

	if this.OutLockCommissionTransactionRecord != nil && this.OutLockAmountTransactionRecord != nil &&
		this.OutLockCommissionTransactionRecord.ProcessType != this.OutLockAmountTransactionRecord.ProcessType {
		return errors.Errorf("ProcessType not equals OutLockCommissionTransactionRecord %s and OutLockAmountTransactionRecord %s",
			this.OutLockCommissionTransactionRecord.ProcessType,
			this.OutLockAmountTransactionRecord.ProcessType)
	}

	return nil
}

func (lockTransactionRecordInfo LockTransactionRecordInfo) Validate() error {

	err := lockTransactionRecordInfo.ValidateOperationID()
	if err != nil {
		return errors.New("ValidateOperationID" + err.Error())
	}
	lockTransactionRecordInfo.ValidateProcessType()
	if err != nil {
		return errors.New("ValidateProcessType " + err.Error())
	}

	outLockedAmountTransactionRecord := lockTransactionRecordInfo.OutLockAmountTransactionRecord
	inLockedAmountTransactionRecord := lockTransactionRecordInfo.InLockAmountTransactionRecord
	outLockedCommissionTransactionRecord := lockTransactionRecordInfo.OutLockCommissionTransactionRecord
	inLockedCommissionTransactionRecord := lockTransactionRecordInfo.InLockCommissionTransactionRecord

	err = CheckTransactionRecord(outLockedAmountTransactionRecord)
	if err != nil {
		return errors.New("out locked amount " + err.Error())
	}

	err = CheckTransactionRecord(inLockedAmountTransactionRecord)
	if err != nil {
		return errors.New("in locked amount " + err.Error())
	}

	if outLockedAmountTransactionRecord.ProcessType != TransferReceiverCommission_ProcessType &&
		outLockedAmountTransactionRecord.ProcessType != Move_ProcessType {
		err = CheckTransactionRecord(outLockedCommissionTransactionRecord)
		if err != nil {
			return errors.New("out locked commission " + err.Error())
		}

		err = CheckTransactionRecord(inLockedCommissionTransactionRecord)
		if err != nil {
			return errors.New("in locked commission " + err.Error())
		}
	}

	return nil
}

// check TransactionRecord
func CheckTransactionRecord(record *TransactionRecord) error {
	if record == nil {
		return errors.New("transaction record can't be nil")
	}

	err := record.Validate()
	if err != nil {
		return err
	}

	return nil
}

// AccountBalance denote balance of the
// account for given currency id
type AccountBalance struct {
	UserAccountInfo accountCC.UserAccountInfo `json:"user_account_info,omitempty"`
	Balance         nums.PlainCurrencyNumber  `json:"balance"`
}

// currencyTypeKeys
type CurrencyTypeKeys map[string]map[string]bool

// LastLoadKeys map of the recent currency loaded keys
type LastLoadKeys struct {
	MapCurrencyTypeKeys CurrencyTypeKeys `json:"keys"`
}

// ImportCryptocurrencyResult encapsulates results of the
// crypto currency load transaction, namely depositing the
// funds on the user account
type LoadCryptocurrencyResultList struct {
	LoadCryptocurrencyResultList []LoadCryptocurrencyResult `json:"load_cryptocurrency_result_list"`
}

type ResultStatus string

const (
	Ok    ResultStatus = "OK"
	Error ResultStatus = "ERROR"
)

// CryptocurrencyResult encapsulates results
// of cryptocurrency import results
type LoadCryptocurrencyResult struct {
	ExternalTransactionId string               `json:"external_transaction_id"`
	TransactionRecordList []*TransactionRecord `json:"transaction_record_list"`
	Status                ResultStatus         `json:"status"`
	CommonMessage         string               `json:"common_message"`
	DetailMessage         string               `json:"detail_message"`
}

// queryKey document key
type QueryKey struct {
	Key           string `json:"key,omitempty"`
	OperationID   string `json:"operation_id,omitempty"`
	DependencyKey string `json:"dependency_key,omitempty"`
}

// selectTx use selector to query for transaction record from the state
//func (q QueryKey) selectTx(adapter chaincode.StubInterfaceAdapter) (*TransactionRecord, error) {
//	lockedTxQuery, err := chaincode.GetSelector(q)
//	if err != nil {
//		return nil, err
//	}
//
//	// Locked funds transaction record
//	var lockedFundsRecord TransactionRecord
//	it, err := adapter.GetQueryResult(lockedTxQuery)
//	if err != nil {
//		return nil, errors.Errorf("failed to retreive transaction with locked funds, query %s, due to %s", lockedTxQuery, err)
//	}
//
//	if !it.HasNext() {
//		return nil, errors.Errorf("locked funds was not found, out key %s", q.Key)
//	}
//
//	lockedFundsJSONRecord, err := it.Next()
//	if err != nil {
//		return nil, errors.Errorf("filed to read next record from iterator to get transaction record for operation id %s, due to %s", q.OperationID, err)
//	}
//	// Get locked funds information from the state db
//	err = json.Unmarshal(lockedFundsJSONRecord.Value, &lockedFundsRecord)
//	if err != nil {
//		return nil, errors.Errorf("failed to deserialize transaction record for operation id %s, due to %s", q.OperationID, err)
//	}
//
//	return &lockedFundsRecord, nil
//}

type AccountCurrencyBalanceArg struct {
	UserAccountInfo accountCC.UserAccountInfo `json:"user_account_info,omitempty"`
	CurrencyType    string                    `json:"currency_type,omitempty"`
}

type DepositFundsArg struct {

	// канал
	Channel   string `json:"channel"`

	// смартконтракт
	Chaincode string `json:"chaincode"`

	// UserId для сервисного пользовтеля для ввода средств
	DaemonName        string          `json:"daemon_name"`

	// Номер сессии для ввода средств
	SessionID         string          `json:"session_id"`

	// Тип валюты
	CurrencyType      string          `json:"currency_type"`
}

func Unmarshal_DepositFundsArg(data []byte) (*DepositFundsArg, error) {
	if data == nil {
		return nil, chaincode.GlobalLogger.Error("can't Unmarshal nil data")
	}

	var obj DepositFundsArg
	err := json.Unmarshal(data, &obj)
	return &obj, err
}
